import requests
import semver
from typing import Union
from pprint import pprint


from smithed import weld
from beet import Context, DataPack, ResourcePack


from .dep import VersionedDep, Dep, Source, get_id_slug


"""
Utility functions
"""


def get_filename(dep: VersionedDep, pack_type) -> str:
    return f"{dep.identifier}-{pack_type}.zip"


def get_identifier(dep: VersionedDep) -> str:
    return f"{dep.id}@{dep.source}"


def get_dict_identifier(dep: dict):
    return f'{dep["id"]}@{dep["source"]}#{dep["match"]}'


"""
Working with deps
"""


def parse_deps(deps: list[dict]) -> dict[Dep, list[str]]:
    result = {}
    for d in deps:
        id, slug = get_id_slug(d["id"], Source(d["source"]))
        dep = Dep(id=id, source=Source(d["source"]), slug=slug)
        match = d["match"].split(",")
        match = [m.strip() for m in match]
        result[dep] = match
    return result


def resolve_deps(deps: dict[Dep, list[str]], include_prerelease: bool) -> list[VersionedDep]:
    new_deps = dict()
    for dep in deps:
        # sort versions by semver
        versions = sorted(
            dep.get_versions(),
            key=lambda v: semver.VersionInfo.parse(v.version),
            reverse=True,
        )
        if not include_prerelease:
            versions = [
                v
                for v in versions
                if not semver.VersionInfo.parse(v.version).prerelease
            ]
            versions = [
                v 
                for v in versions 
                if not semver.VersionInfo.parse(v.version).build
            ]

        # filter versions by match
        versions = [v for v in versions if all([semver.match(v.version, deps[dep][i]) for i in range(len(deps[dep]))])]
        # append newest version
        new_deps[get_identifier(versions[0])] = versions[0]
        # append dependencies
        add_deps(versions[0], new_deps)
    return new_deps.values()


def add_deps(dep: VersionedDep, deps: dict[str, VersionedDep]):
    for d in dep.dependencies:
        id_d = get_identifier(d)
        if id_d not in deps:
            deps[id_d] = d
        elif semver.VersionInfo.parse(deps[id_d].version) < semver.VersionInfo.parse(
            d.version
        ):
            deps[id_d] = d
        if len(d.dependencies) > 0:
            add_deps(d, deps)


"""
Caching
"""


def load_deps(
    ctx: Context, files: list[str], clean_load_tag: bool, enable_weld_merging: bool
):
    if enable_weld_merging:
        weld.toolchain.main.weld(ctx)
    cache_dir = ctx.cache.path / "weld_deps"

    for file in files:
        if file.endswith("dp.zip"):
            data = DataPack(zipfile=cache_dir / file)
            data = clean_pack(ctx, data, clean_load_tag)
            ctx.data.merge(data)
        elif file.endswith("rp.zip"):
            resource = ResourcePack(zipfile=cache_dir / file)
            resource = clean_pack(ctx, resource, clean_load_tag)
            ctx.assets.merge(resource)


def cache_deps(ctx: Context, deps: list[VersionedDep]):
    cache_dir = ctx.cache.path / "weld_deps"
    cache_dir.mkdir(exist_ok=True)

    result = []
    result_urls = []

    for dep in deps:
        datapack_id = get_filename(dep, "dp")
        datapack_path = cache_dir / datapack_id
        resourcepack_id = get_filename(dep, "rp")
        resourcepack_path = cache_dir / resourcepack_id

        if dep.datapack_download_url and dep.datapack_download_url.startswith(
            "https://"
        ):
            if not datapack_path.exists():
                datapack = dep.get_datapack()
                datapack_path.write_bytes(datapack)
            result.append(get_filename(dep, "dp"))
            result_urls.append(dep.datapack_download_url)
        if dep.resourcepack_download_url and dep.resourcepack_download_url.startswith(
            "https://"
        ):
            if not resourcepack_path.exists():
                resourcepack = dep.get_resourcepack()
                resourcepack_path.write_bytes(resourcepack)
            result.append(get_filename(dep, "rp"))
            result_urls.append(dep.resourcepack_download_url)
    return result, result_urls


def download_from_urls(ctx: Context, urls: list[str], files: list[str]):
    cache_dir = ctx.cache.path / "weld_deps"
    cache_dir.mkdir(exist_ok=True)

    for url, file in zip(urls, files):
        path = cache_dir / file
        if not path.exists():
            path.write_bytes(requests.get(url).content)


def clean_pack(
    ctx: Context, pack: Union[DataPack, ResourcePack], clean_load_tag: bool
) -> Union[DataPack, ResourcePack]:
    if "pack.png" in pack.extra:
        del pack.extra["pack.png"]
    if (
        isinstance(pack, DataPack)
        and "load:load" in pack.function_tags
        and clean_load_tag
    ):
        del pack.function_tags["load:load"]
    return pack



from enum import Enum
from dataclasses import dataclass
import requests
import semver
from typing import Dict, List
import itertools
from pydantic import BaseModel
from pprint import pprint

CACHED_VERSIONS: dict[str, List["VersionedDep"]] = {}

UID_MAP: dict[str, str] = {}
SLUG_MAP: dict[str, str] = {}


def get_id_slug(id: str, source: "Source") -> str:
    if id in UID_MAP:
        return UID_MAP[id], SLUG_MAP[id]
    elif source == Source.SMITHED:
        r = requests.get(f"https://api.smithed.dev/v2/packs/{id}/meta")
        r.raise_for_status()
        d = r.json()
        raw_id = d["rawId"]
        uid = d["docId"]
        UID_MAP[raw_id] = uid
        UID_MAP[uid] = uid
        SLUG_MAP[raw_id] = raw_id
        SLUG_MAP[uid] = raw_id
        return uid, raw_id
    elif source == Source.MODRINTH:
        r = requests.get(f"https://api.modrinth.com/v2/project/{id}")
        r.raise_for_status()
        d = r.json()
        slug = d["slug"]
        uid = d["id"]
        UID_MAP[slug] = uid
        UID_MAP[uid] = uid
        SLUG_MAP[slug] = slug
        SLUG_MAP[uid] = slug
        return uid, slug
    else:
        raise ValueError("Invalid source")


class Source(Enum):
    SMITHED = "smithed"
    MODRINTH = "modrinth"


class Dep(BaseModel):
    source: Source
    id: str
    slug: str

    @property
    def identifier(self) -> str:
        return f"{self.id}@{self.source}"

    def __hash__(self) -> int:
        return hash(self.identifier)

    @property
    def versions_url(self) -> str:
        if self.source == Source.SMITHED:
            return f"https://api.smithed.dev/v2/packs/{self.id}/versions"
        elif self.source == Source.MODRINTH:
            return f"https://api.modrinth.com/v2/project/{self.id}/version"

    def get_versions(self) -> List["VersionedDep"]:
        if self.identifier in CACHED_VERSIONS:
            return CACHED_VERSIONS[self.identifier]
        r = requests.get(self.versions_url)
        r.raise_for_status()
        versions = r.json()

        l_versions = []
        if self.source == Source.SMITHED:
            for v in versions:
                l_deps = []
                for dep in v["dependencies"]:
                    id, slug = get_id_slug(dep["id"], Source.SMITHED)
                    d_obj = Dep(
                        source=self.source,
                        id=id,
                        slug=slug,
                    )
                    d_obj_versions = d_obj.get_versions()
                    d_obj_version = [
                        v for v in d_obj_versions if v.version == dep["version"]
                    ][0]
                    l_deps.append(d_obj_version)

                v_obj = VersionedDep(
                    id=self.id,
                    slug=self.slug,
                    source=self.source,
                    version=v["name"],
                    dependencies=l_deps,
                    datapack_download_url=v["downloads"]["datapack"]
                    if "datapack" in v["downloads"]
                    else None,
                    resourcepack_download_url=v["downloads"]["resourcepack"]
                    if "resourcepack" in v["downloads"]
                    else None,
                )
                l_versions.append(v_obj)
        elif self.source == Source.MODRINTH:
            for v in versions:
                l_deps = []
                for dep in v["dependencies"]:
                    if dep["dependency_type"] == "optional":
                        continue
                    id, slug = get_id_slug(dep["project_id"], Source.MODRINTH)
                    d_obj = Dep(
                        source=self.source,
                        id=id,
                        slug=slug,
                    )
                    d_obj_versions = d_obj.get_versions()
                    if dep["version_id"] is None:
                        continue
                    d_obj_version = [
                        v for v in d_obj_versions if v.version_id == dep["version_id"]
                    ][0]
                    l_deps.append(d_obj_version)

                dp_url, rp_url = None, None
                for f in v["files"]:
                    if f["primary"]:
                        dp_url = f["url"]
                    elif f["file_type"] in [
                        "required-resource-pack",
                        "optional-resource-pack",
                    ]:
                        rp_url = f["url"]

                v_obj = VersionedDep(
                    id=self.id,
                    slug=self.slug,
                    source=self.source,
                    version=v["version_number"],
                    version_id=v["id"],
                    dependencies=l_deps,
                    datapack_download_url=dp_url,
                    resourcepack_download_url=rp_url,
                )
                l_versions.append(v_obj)
        CACHED_VERSIONS[self.identifier] = l_versions
        return l_versions


class VersionedDep(Dep):
    version: str
    dependencies: list["VersionedDep"]
    datapack_download_url: str | None
    resourcepack_download_url: str | None
    version_id: str | None = None

    @property
    def identifier(self) -> str:
        return f"{self.id}@{self.source}#{self.version}"

    def __hash__(self) -> int:
        return hash(self.identifier)

    def get_datapack(self) -> bytes:
        if self.datapack_download_url is None:
            raise ValueError("No datapack download url")
        r = requests.get(self.datapack_download_url)
        r.raise_for_status()
        return r.content

    def get_resourcepack(self) -> bytes:
        if self.resourcepack_download_url is None:
            raise ValueError("No resourcepack download url")
        r = requests.get(self.resourcepack_download_url)
        r.raise_for_status()
        return r.content



from beet import Context
from pprint import pprint
from .utils import (
    get_dict_identifier,
    parse_deps,
    resolve_deps,
    cache_deps,
    download_from_urls,
    load_deps,
)


def beet_default(ctx: Context):

    if "weld_deps" not in ctx.meta:
        return
    if ctx.meta["weld_deps"].get("enabled", True) == False:
        return


    enable_weld_merging = ctx.meta["weld_deps"].get("enable_weld_merging", True)
    clean_load_tag = ctx.meta["weld_deps"].get("clean_load_tag", True)
    include_prerelease = ctx.meta["weld_deps"].get("include_prerelease", False)
    params = {
        "clean_load_tag": clean_load_tag,
        "enable_weld_merging": enable_weld_merging,
        "include_prerelease": include_prerelease,
    }

    deps = ctx.meta.get("weld_deps", {}).get("deps", [])
    deps = [d for d in deps if d["source"] != "integrated"]
    if len(deps) == 0:
        return
    # get a hash of the deps
    deps_hash = set()
    [deps_hash.add(get_dict_identifier(d)) for d in deps]

    # check if deps are cached
    if "weld_deps" in ctx.cache.json:
        cache = ctx.cache.json["weld_deps"]
        if "deps" in cache:
            if set(cache["deps"]) == deps_hash and cache["params"] == params:
                download_from_urls(ctx, cache["urls"], cache["files"])
                load_deps(ctx, cache["files"], clean_load_tag, enable_weld_merging)
                return

    deps = parse_deps(deps)

    deps = resolve_deps(deps, include_prerelease)

    # cache deps
    files, urls = cache_deps(ctx, deps)

    ctx.cache.json["weld_deps"] = {
        "deps": list(deps_hash),
        "params": params,
        "files": list(files),
        "urls": list(urls),
    }

    ctx.cache.json["weld_deps_installed"] = {dep.slug: dep.version for dep in deps}

    load_deps(ctx, files, clean_load_tag, enable_weld_merging)
